/*  Pcsx2 - Pc Ps2 Emulator
 *  Copyright (C) 2002-2003  Pcsx2 Team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/stat.h>

#include "PsxCommon.h"

#ifdef __WIN32__
#pragma warning(disable:4244)
#endif
#ifdef __VCNET2005__
#pragma warning(disable:4996)//deprecate api annoying messages in vsnet2005beta1
#endif

// *** FOR WORKS ON PADS AND MEMORY CARDS *****

static unsigned char buf[256];
static unsigned char buf1[] = {0x25, 0x46, 0x01, 0x31, 0x00, 0xA2, 0x11, 0x01, 0xE1};//{0x64, 0x23, 0x2, 0x43, 0x0, 0xa2, 0x11, 0x1, 0xb4}; 
static unsigned char buf2[] = {0xC5, 0xBD, 0x66, 0x00, 0x59, 0x44, 0x01, 0x02, 0x00};//{0x6d, 0x21, 0x30, 0x0, 0x55, 0x19, 0x2, 0x2, 0x30};
static unsigned char buf4[] = {0x02, 0x9A, 0x9E, 0x06, 0x6D, 0x3C, 0xF0, 0x7E, 0xDF};//{0xa3, 0x5d, 0x2f, 0xa2, 0xd8, 0x7c, 0x5b, 0x35, 0xb9};
static unsigned char buff[] = {0xC2, 0x39, 0x6F, 0x27, 0xC8, 0xDF, 0x2A, 0x23, 0xAD};//{0xf3, 0x9b, 0x32, 0x87, 0x31, 0xda, 0x9d, 0x10, 0xbb};
static unsigned char buf11[] = {0xA8, 0x42, 0x5D, 0x87, 0x65, 0x32, 0x6F, 0xE8, 0xE0};//{0x39, 0xd2, 0xb9, 0x5c, 0xcf, 0x31, 0x2d, 0x23, 0xfe};
static unsigned char buf13[] = {0x46, 0x31, 0xFC, 0x97, 0xA8, 0x6D, 0xE2, 0x12, 0x29};//{0x20, 0x2e, 0xd7, 0x99, 0x92, 0x29, 0x4a, 0x12, 0xa3};


unsigned char cardh[4] = { 0xFF, 0xFF, 0x5a, 0x5d };
struct mc_command_0x26_tag mc_command_0x26=
	{'+', 512, 16, 0x4000, 0x52, 'Z'};//sizeof()==11

//static unsigned short StatReg = 0x002b;
// Transfer(?) Ready and the Buffer is Empty
unsigned short StatReg = TX_RDY | TX_EMPTY;
unsigned short ModeReg;
unsigned short CtrlReg;
unsigned short BaudReg;

static unsigned long bufcount;
static unsigned long parp;
static unsigned long mcdst,rdwr;
static unsigned char adrH,adrL;
static unsigned long padst;
static unsigned long mtapst;
unsigned long packetsize = 0;

// clk cycle byte
// 4us * 8bits = ((PSXCLK / 1000000) * 32) / BIAS; (linuzappz)
#define SIO_INT() PSX_INT(16, 200); /*270;*/

void _ReadMcd(char *data, u32 adr, int size) {
	ReadMcd(CtrlReg&0x2000?2:1, data, adr, size);
}

void _SaveMcd(char *data, u32 adr, int size) {
	SaveMcd(CtrlReg&0x2000?2:1, data, adr, size);
}

unsigned char xor(unsigned char *buf, unsigned int length){
	register unsigned char i, x;

	for (x=0, i=0; i<length; i++)	x ^= buf[i];
	return x & 0xFF;
}

unsigned char sioRead8() {
	unsigned char ret = 0xFF;

	if ((StatReg & RX_RDY)/* && (CtrlReg & RX_PERM)*/) {
//		StatReg &= ~RX_OVERRUN;
		ret = buf[parp];
		if (parp == bufcount) {
			StatReg &= ~RX_RDY;		// Receive is not Ready now?
			StatReg |= TX_EMPTY;	// Buffer is Empty
			if (mcdst == 5) {
				mcdst = 0;
				if (rdwr == 2) {
					_SaveMcd(&buf[1], (adrL | (adrH << 8)) * 128, 128);
				}
			}
			if (padst == 2) padst = 0;
			if (mcdst == 1) {
				mcdst = 2;
				StatReg&= ~TX_EMPTY;
				StatReg|= RX_RDY;
			}
		}
	}

#ifdef PAD_LOG
	PAD_LOG("sio read8 ;ret = %x\n", ret);
#endif
	return ret;
}

void sioWrite8(unsigned char value) {
	u32 i = 0;
#ifdef PAD_LOG
	PAD_LOG("sio write8 %x\n", value);
#endif
	switch (padst) {
		case 1: SIO_INT();
			if ((value&0x40) == 0x40) {
				padst = 2; parp = 1;
				switch (CtrlReg&0x2002) {
					case 0x0002:
						packetsize ++;	// Total packet size sent
						buf[parp] = PAD1poll(value);
						break;
					case 0x2002:
						packetsize ++;	// Total packet size sent
						buf[parp] = PAD2poll(value);
						break;
				}
				if (!(buf[parp] & 0x0f)) {
					bufcount = 2 + 32;
				} else {
					bufcount = 2 + (buf[parp] & 0x0f) * 2;
				}
			}
			else padst = 0;
			return;
		case 2:
			parp++;
			switch (CtrlReg&0x2002) {
				case 0x0002: packetsize ++; buf[parp] = PAD1poll(value); break;
				case 0x2002: packetsize ++; buf[parp] = PAD2poll(value); break;
			}
			if (parp == bufcount) { padst = 0; return; }
			SIO_INT();
			return;
	}

	switch (mcdst) {static u8 mc_command=0x00, i; static u32 lastsector=0, sector=0, k=0;
					static u8 terminator='Z'; static u8 mode=0;
		case 1:
			packetsize++;
			SIO_INT();
			if (rdwr) { parp++; return; }
			parp = 1;
			buf[0]=buf[1]=0xFF;
			switch (value) {
			case 0x11: 
			case 0x12: 
			case 0x81:
				bufcount =  8; 
				memset(buf, 0xFF, bufcount+1);
				buf[3] = terminator;
				buf[2] = '+';
				mcdst = 99; 
				sio2.packet.recvVal3 = 0x8c;
				if(value == 0x81)
				{
					if(mc_command==0x42)
						sio2.packet.recvVal1 = 0x1600; // Writing
					else if(mc_command==0x43) sio2.packet.recvVal1 = 0x1700; // Reading
				}
				break;
			case 0x21: 
			case 0x22: 
			case 0x23: 
                bufcount =  8; mcdst = 99; sector=0;k=0;
				memset(buf, 0xFF, bufcount+1);
				sio2.packet.recvVal3 = 0x8c; 
				buf[bufcount]=terminator;
				buf[bufcount-1]='+';
                break;
			case 0x24:											break;//! 
			case 0x25:											break;//!
			case 0x26: 
				bufcount = 12; mcdst = 99; sio2.packet.recvVal3 = 0x83;	
				memset(buf, 0xFF, bufcount+1);
				memcpy(&buf[2], &mc_command_0x26, sizeof(mc_command_0x26));
				buf[bufcount]=terminator;
				break;
			case 0x27: 
			case 0x28: 
			case 0xBF:
				bufcount =  8; mcdst = 99; sio2.packet.recvVal3 = 0x8b;
				memset(buf, 0xFF, bufcount+1);
				buf[4]=terminator;
				buf[3]='+';
				break;
			case 0x42: 
			case 0x43: 
			case 0x82:
				if(value==0x82 && lastsector==sector) mode = 2;
				if(value==0x42) mode = 0;
				if(value==0x43)	lastsector = sector;
				
				bufcount =133; mcdst = 99;
				memset(buf, 0xFF, bufcount+1);
				buf[bufcount]=terminator;
				buf[bufcount-1]='+';
				break;
			case 0xf0: 		//don't handle this here, see below		  //__card_auth_00_03_05_08_0a_0c_0e_10_12_14
			case 0xf1:												  //__card_auth_00_03_05_08_0a_0c_0e_10_12_14
			case 0xf2:				  
				mcdst = 99;	
				break;//__card_auth_00_03_05_08_0a_0c_0e_10_12_14
			case 0xf3: //__card_auth_0x60_F3
			case 0xf7: 
				bufcount = 8; mcdst = 99;
				memset(buf, 0xFF, bufcount+1);
				buf[4]=terminator;
				buf[3]='+';
				break;//__card_auth_key_change_F7(sendBuf[2])
			case 0x52: rdwr = 1; memset(buf, 0xFF, bufcount+1); buf[bufcount]=terminator; buf[bufcount-1]='+';	break;
			case 0x57: rdwr = 2; memset(buf, 0xFF, bufcount+1); buf[bufcount]=terminator; buf[bufcount-1]='+';	break;
			default: mcdst = 0;
					memset(buf, 0xFF, bufcount+1);
					buf[bufcount]=terminator; buf[bufcount-1]='+';	
#ifdef MEMCARDS_LOG
					MEMCARDS_LOG("Unknown MC(%d) command 0x%02X\n", ((CtrlReg&0x2000)>>13)+1, value);
#endif	
			}
			mc_command=value;
			if (mc_command!=0x21 && mc_command!=0x22 && mc_command!=0x23 &&
			   (mc_command!=0xF0) && (mc_command!=0xF1) && (mc_command!=0xF2)) {
#ifdef MEMCARDS_LOG
				MEMCARDS_LOG("MC(%d) command 0x%02X\n", ((CtrlReg&0x2000)>>13)+1, value);
#endif
			}
			return;
		case 99://as is...
			packetsize++;
			parp++;
			switch(mc_command)
			{
			case 0x21:
			case 0x22:
			case 0x23:
                if (parp==2)sector|=(value & 0xFF)<< 0;
				if (parp==3)sector|=(value & 0xFF)<< 8;
				if (parp==4)sector|=(value & 0xFF)<<16;
				if (parp==5)sector|=(value & 0xFF)<<24;
				if (parp==6)
				{
#ifdef MEMCARDS_LOG
                 MEMCARDS_LOG("MC(%d) command 0x%02X sector 0x%04X\n",
								((CtrlReg&0x2000)>>13)+1, mc_command, sector);
#endif
				}
				break;
			case 0x27:
				if(parp==2)	{
					terminator = value;
					buf[4] = value;
				}
				break;
			case 0x28:
				if(parp == 2)
				{
					buf[2] = '+';
					if(value == 0) 
					{
						buf[4] = 0xFF;
						buf[3] = terminator;
					}else{
						buf[4] = 0x5A;
						buf[3] = terminator;
					}
				}	
				break;
			case 0x42:
				if (parp==2)
				{
					bufcount=5+value;
					memset(buf, 0xFF, bufcount+1);
					buf[bufcount-1]='+';
					buf[bufcount]=terminator;
				}else if ((parp>2) && (parp<bufcount-2))
					buf[parp]=value;
				else if (parp==bufcount-2)
				{
					if (xor(&buf[3], bufcount-5)==value)
					{
                        _SaveMcd(&buf[3], (512+16)*sector+k, bufcount-5);
						buf[bufcount-1]=value;
						k+=bufcount-5;
					}else {
#ifdef MEMCARDS_LOG
						MEMCARDS_LOG("MC(%d) write XOR value error 0x%02X != ^0x%02X\n",
							((CtrlReg&0x2000)>>13)+1, value, xor(&buf[3], bufcount-5));
#endif
					}
				}
				break;
			case 0x43:
				if (parp==2){
						bufcount=value+5;
						buf[3]='+';
						_ReadMcd(&buf[4], (512+16)*sector+k, value);
						if(mode==2)
						{
							int j;
							for(j=0; j < value; j++)
								buf[4+j] = ~buf[4+j];
						}
						k+=value;
						buf[bufcount-1]=xor(&buf[4], value);
						buf[bufcount]=terminator;
				}
				break;
			case 0x82:
				if(parp==2)
				{
					buf[2]='+';
					buf[3]=terminator;
				}
				break;
			case 0xF0:
				if (parp==2)
				{
					u32 flag=0;
#ifdef MEMCARDS_LOG
					MEMCARDS_LOG("MC(%d) command 0xF0:0x%02X\n", ((CtrlReg&0x2000)>>13)+1, value);
#endif
					switch(value){
					case  1:
						flag = 0;
						bufcount=13;
						memset(buf, 0xFF, bufcount+1);
						memcpy(&buf[4],buf1,9);//xor value for OUT data
						buf[3]='+';
						break;//OUT
					case  2:
						flag = 0;
						bufcount=13;
						memset(buf, 0xFF, bufcount+1);
						memcpy(&buf[4],buf2,9);//xor value for OUT data
						buf[3]='+';
						break;//OUT
					case  4:
						flag = 0;
						bufcount=13;
						memset(buf, 0xFF, bufcount+1);
						memcpy(&buf[4],buf4,9);//xor value for OUT data
						buf[3]='+';
						break;//OUT
					case  6:
					case  7:
					case 11:
						flag=0;
						bufcount=13;
						memset(buf, 0xFF, bufcount+1);
						buf[bufcount-1]='+';
						break;//IN
					case 15:
						flag = 0;
						bufcount=13;
						memset(buf, 0xFF, bufcount+1);
						memcpy(&buf[4],buff,9);//xor value for OUT data
						buf[3]='+';
						break;//OUT
					case 17:
						flag = 0;
						bufcount=13;
						memset(buf, 0xFF, bufcount+1);
						memcpy(&buf[4],buf11,9);//xor value for OUT data
						buf[3]='+';
						break;//OUT
					case 19:
						flag = 0;
						bufcount=13;
						memset(buf, 0xFF, bufcount+1);
						memcpy(&buf[4],buf13,9);//xor value for OUT data
						buf[3]='+';
						break;//OUT
					default:
						bufcount=4;
						memset(buf, 0xFF, bufcount+1);
						buf[bufcount-1]='+';
					}
					
					if ((bufcount==13) && flag){
						buf[bufcount-1] = 0;//xor value for OUT data
						buf[3]='+';
					}
					buf[bufcount]=terminator;
				}
				break;
			}
			if (bufcount<=parp)	mcdst = 0;
			return;
	}

	switch (mtapst) {
		case 0x1:
			packetsize++;
			parp = 1;
			SIO_INT();
			switch(value) {
			case 0x12:  mtapst = 2; bufcount = 5;	break;
			case 0x13:  mtapst = 2; bufcount = 5;	break;
			case 0x21:	mtapst = 2; bufcount = 6;	break;
			}
			buf[bufcount]='Z';
			buf[bufcount-1]='+';
			return;
		case 0x2:
			packetsize++;
			parp++;
            if (bufcount<=parp)	mcdst = 0;
			SIO_INT();
			return;
	}

	switch (value) {
		case 0x01: // start pad
			StatReg &= ~TX_EMPTY;	// Now the Buffer is not empty
			StatReg |= RX_RDY;		// Transfer is Ready

			switch (CtrlReg&0x2002) {
				case 0x0002: buf[0] = PAD1startPoll(1); break;
				case 0x2002: buf[0] = PAD2startPoll(2); break;
			}

			bufcount = 2;
			parp = 0;
			padst = 1;
			packetsize = 1; // Count this one too ! :P
			sio2.packet.recvVal1 = 0x1100; // Pad is present
			SIO_INT();
			return;

		case 0x21: // start mtap
			StatReg &= ~TX_EMPTY;	// Now the Buffer is not empty
			StatReg |= RX_RDY;		// Transfer is Ready
			parp = 0;
			packetsize = 1; // Count this one too ! :P
			mtapst = 1;
			sio2.packet.recvVal1 = 0x1D100; // Mtap is not connected :)
			SIO_INT();
			return;

		case 0x61: // start remote control sensor
			StatReg &= ~TX_EMPTY;	// Now the Buffer is not empty
			StatReg |= RX_RDY;		// Transfer is Ready
			parp = 0;
			packetsize = 1; // Count this one too ! :P
			sio2.packet.recvVal1 = 0x1100; // Pad is present
			SIO_INT();
			return;

		case 0x81: // start memcard
			StatReg &= ~TX_EMPTY;
			StatReg |= RX_RDY;
			memcpy(buf, cardh, 4);
			parp = 0;
			bufcount = 3;
			mcdst = 1;
			packetsize = 1;
			rdwr = 0;
			sio2.packet.recvVal1 = 0x1100; // Memcard1 is present
			SIO_INT();
			return;
	}
}

void sioWriteCtrl16(unsigned short value) {
	CtrlReg = value & ~RESET_ERR;
	if (value & RESET_ERR) StatReg &= ~IRQ;
	if ((CtrlReg & SIO_RESET) || (!CtrlReg)) {
		mtapst = 0; padst = 0; mcdst = 0; parp = 0;
		StatReg = TX_RDY | TX_EMPTY;
		psxRegs.interrupt&= ~(1<<16);
	}
}

int  sioInterrupt() {
#ifdef PAD_LOG
	PAD_LOG("Sio Interrupt\n");
#endif
	StatReg|= IRQ;
	psxHu32(0x1070)|=0x80;
	return 1;
}

FILE *LoadMcd(int mcd) {
	char str[256];
	FILE *f;

	if (mcd == 1) {
		strcpy(str, Config.Mcd1);
	} else {
		strcpy(str, Config.Mcd2);
	}
	if (*str == 0) sprintf(str, "memcards/Mcd00%d.ps2", mcd);
	f = fopen(str, "r+b");
	if (f == NULL) {
		CreateMcd(str);
		f = fopen(str, "r+b");
	}
	if (f == NULL) {
		SysMessage (_("Failed loading MemCard %s"), str); return NULL;
	}

	return f;
}

void SeekMcd(FILE *f, u32 adr) {
	u32 size;

	fseek(f, 0, SEEK_END); size = ftell(f);
	if (size == MCD_SIZE + 64)
		fseek(f, adr + 64, SEEK_SET);
	else if (size == MCD_SIZE + 3904)
		fseek(f, adr + 3904, SEEK_SET);
	else
		fseek(f, adr, SEEK_SET);
}

void ReadMcd(int mcd, char *data, u32 adr, int size) {
	FILE *f = LoadMcd(mcd);
	if (f == NULL) {
		memset(data, 0, size);
		return;
	}
	SeekMcd(f, adr);
	fread(data, 1, size, f);
	fclose(f);
}

void SaveMcd(int mcd, char *data, u32 adr, int size) {
	FILE *f = LoadMcd(mcd);
	if (f == NULL) {
		return;
	}
	SeekMcd(f, adr);
	fwrite(data, 1, size, f);
	fclose(f);
}

void CreateMcd(char *mcd) {
	FILE *fp;	
	int i=0, j=0;
	int enc[16] = {0x77,0x7f,0x7f,0x77,0x7f,0x7f,0x77,0x7f,0x7f,0x77,0x7f,0x7f,0,0,0,0};

	fp = fopen(mcd, "wb");
	if (fp == NULL) return;
	for(i=0; i < 16384; i++) 
	{
		for(j=0; j < 128; j++) fputc(0x00,fp);
		for(j=0; j < 128; j++) fputc(0x00,fp);
		for(j=0; j < 128; j++) fputc(0x00,fp);
		for(j=0; j < 128; j++) fputc(0x00,fp);
		for(j=0; j < 16; j++) fputc(enc[j],fp);
	}
	fclose(fp);
}

int sioFreeze(gzFile f, int Mode) {
	gzfreezel(buf);
	gzfreezel(&StatReg);
	gzfreezel(&ModeReg);
	gzfreezel(&CtrlReg);
	gzfreezel(&BaudReg);
	gzfreezel(&bufcount);
	gzfreezel(&parp);
	gzfreezel(&mcdst);
	gzfreezel(&rdwr);
	gzfreezel(&adrH);
	gzfreezel(&adrL);
	gzfreezel(&padst);

	return 0;
}


