
#include "ix86.h"

/********************/
/* SSE instructions */
/********************/

#define SSEMtoR( code, overb ) \
	write16( code ); \
	ModRM( 0, to, DISP32 ); \
	write32( MEMADDR(from, 4 + overb) ); \

#define SSERtoM( code, overb ) \
	write16( code ); \
	ModRM( 0, from, DISP32 ); \
	write32( MEMADDR(to, 4 + overb) ); \

#define SSE_SS_MtoR( code1, code2 ) \
	write16( code1 ); \
	write8( code2 ); \
	ModRM( 0, to, DISP32 ); \
	write32( MEMADDR(from, 4) ); \

#define SSE_SS_RtoM( code1, code2 ) \
	write16( code1 ); \
	write8( code2 ); \
	ModRM( 0, from, DISP32 ); \
	write32( MEMADDR(to, 4) ); \

#define SSERtoR( code ) \
	write16( code ); \
	ModRM( 3, to, from );

#define SSEMtoR66( code ) \
	write8( 0x66 ); \
	SSEMtoR( code, 0 );

#define SSERtoM66( code ) \
	write8( 0x66 ); \
	SSERtoM( code, 0 );

#define SSERtoR66( code ) \
	write8( 0x66 ); \
	SSERtoR( code );

#define SSE_SS_RtoR( code1, code2 ) \
	write16( code1 ); \
	write8( code2 ); \
	ModRM( 3, to, from );

#define CMPPSMtoR( op ) \
   SSEMtoR( 0xc20f, 1 ); \
   write8( op );

#define CMPPSRtoR( op ) \
   SSERtoR( 0xc20f ); \
   write8( op );



void MOVAPSMtoR( x86SSERegType to, u32 from )          { SSEMtoR( 0x280f, 0 ); }
void MOVAPSRtoM( u32 to, x86SSERegType from )          { SSERtoM( 0x290f, 0 ); }
void MOVAPSRtoR( x86SSERegType to, x86SSERegType from )     { SSERtoR( 0x290f ); }

/* movups [r32][r32*scale] to xmm1 */
void MOVUPSRmStoR( x86SSERegType to, x86IntRegType from, x86IntRegType from2, int scale )
{
	write16( 0x100f );
	ModRM( 0, to, 0x4 );
	SibSB( scale, from2, from );
}

/* movups xmm1 to [r32][r32*scale] */
void MOVUPSRtoRmS( x86SSERegType to, x86IntRegType from, x86IntRegType from2, int scale )
{
	write16( 0x110f );
	ModRM( 0, to, 0x4 );
	SibSB( scale, from2, from );
}

/* movups [r32] to r32 */
void MOVUPSRmtoR( x86IntRegType to, x86IntRegType from ) 
{
	write16( 0x100f );
	ModRM( 0, to, from );
}

/* movups r32 to [r32] */
void MOVUPSRtoRm( x86IntRegType to, x86IntRegType from ) 
{
	write16( 0x110f );
	ModRM( 0, from, to );
}

/* movaps [r32][r32*scale] to xmm1 */
void MOVAPSRmStoR( x86SSERegType to, x86IntRegType from, x86IntRegType from2, int scale )
{
	write16( 0x100f );
	ModRM( 0, to, 0x4 );
	SibSB( scale, from2, from );
}

/* movaps xmm1 to [r32][r32*scale] */
void MOVAPSRtoRmS( x86SSERegType to, x86IntRegType from, x86IntRegType from2, int scale )
{
	write16( 0x110f );
	ModRM( 0, to, 0x4 );
	SibSB( scale, from2, from );
}

/* movaps [r32] to r32 */
void MOVAPSRmtoR( x86IntRegType to, x86IntRegType from ) 
{
	write16( 0x280f );
	ModRM( 0, to, from );
}

/* movaps r32 to [r32] */
void MOVAPSRtoRm( x86IntRegType to, x86IntRegType from ) 
{
	write16( 0x290f );
	ModRM( 0, from, to );
}

void MOVSSMtoR( x86SSERegType to, u32 from )           { SSE_SS_MtoR( 0x0ff3, 0x10 ); }
void MOVSSRtoM( u32 to, x86SSERegType from )           { SSE_SS_RtoM( 0x0ff3, 0x11 ); }
void MOVSSRtoR( x86SSERegType to, x86SSERegType from )      { SSE_SS_RtoR( 0x0ff3, 0x10 ); }

void MOVLPSMtoR( x86SSERegType to, u32 from )          { SSEMtoR( 0x120f, 0 ); }
void MOVLPSRtoM( u32 to, x86SSERegType from )          { SSERtoM( 0x130f, 0 ); }

void MOVHPSMtoR( x86SSERegType to, u32 from )          { SSEMtoR( 0x160f, 0 ); }
void MOVHPSRtoM( u32 to, x86SSERegType from )          { SSERtoM( 0x170f, 0 ); }

void MOVLHPSRtoR( x86SSERegType to, x86SSERegType from )    { SSERtoR( 0x160f ); }
void MOVHLPSRtoR( x86SSERegType to, x86SSERegType from )    { SSERtoR( 0x120f ); }


void ORPSMtoR( x86SSERegType to, u32 from )            { SSEMtoR( 0x560f, 0 ); }
void ORPSRtoR( x86SSERegType to, x86SSERegType from )       { SSERtoR( 0x560f ); }

void XORPSMtoR( x86SSERegType to, u32 from )           { SSEMtoR( 0x570f, 0 ); }
void XORPSRtoR( x86SSERegType to, x86SSERegType from )      { SSERtoR( 0x570f ); }

void ANDPSMtoR( x86SSERegType to, u32 from )           { SSEMtoR( 0x540f, 0 ); }
void ANDPSRtoR( x86SSERegType to, x86SSERegType from )      { SSERtoR( 0x540f ); }

void ANDNPSMtoR( x86SSERegType to, u32 from )          { SSEMtoR( 0x550f, 0 ); }
void ANDNPSRtoR( x86SSERegType to, x86SSERegType from )     { SSERtoR( 0x550f ); }

void ADDPSMtoR( x86SSERegType to, u32 from )           { SSEMtoR( 0x580f, 0 ); }
void ADDPSRtoR( x86SSERegType to, x86SSERegType from )      { SSERtoR( 0x580f ); }

void SUBPSMtoR( x86SSERegType to, u32 from )           { SSEMtoR( 0x5c0f, 0 ); }
void SUBPSRtoR( x86SSERegType to, x86SSERegType from )      { SSERtoR( 0x5c0f ); }

void MULPSMtoR( x86SSERegType to, u32 from )           { SSEMtoR( 0x590f, 0 ); }
void MULPSRtoR( x86SSERegType to, x86SSERegType from )      { SSERtoR( 0x590f ); }

void RCPPSRtoR( x86SSERegType to, x86SSERegType from )      { SSERtoR( 0x530f ); }
void RCPPSMtoR( x86SSERegType to, u32 from )           { SSEMtoR( 0x530f, 0 ); }

void CMPEQPSMtoR( x86SSERegType to, u32 from )         { CMPPSMtoR( 0 ); }
void CMPEQPSRtoR( x86SSERegType to, x86SSERegType from )    { CMPPSRtoR( 0 ); }
void CMPLTPSMtoR( x86SSERegType to, u32 from )         { CMPPSMtoR( 1 ); }
void CMPLTPSRtoR( x86SSERegType to, x86SSERegType from )    { CMPPSRtoR( 1 ); }
void CMPLEPSMtoR( x86SSERegType to, u32 from )         { CMPPSMtoR( 2 ); }
void CMPLEPSRtoR( x86SSERegType to, x86SSERegType from )    { CMPPSRtoR( 2 ); }
void CMPUNORDPSMtoR( x86SSERegType to, u32 from )      { CMPPSMtoR( 3 ); }
void CMPUNORDPSRtoR( x86SSERegType to, x86SSERegType from ) { CMPPSRtoR( 3 ); }
void CMPNEPSMtoR( x86SSERegType to, u32 from )         { CMPPSMtoR( 4 ); }
void CMPNEPSRtoR( x86SSERegType to, x86SSERegType from )    { CMPPSRtoR( 4 ); }
void CMPNLTPSMtoR( x86SSERegType to, u32 from )        { CMPPSMtoR( 5 ); }
void CMPNLTPSRtoR( x86SSERegType to, x86SSERegType from )   { CMPPSRtoR( 5 ); }
void CMPNLEPSMtoR( x86SSERegType to, u32 from )        { CMPPSMtoR( 6 ); }
void CMPNLEPSRtoR( x86SSERegType to, x86SSERegType from )   { CMPPSRtoR( 6 ); }
void CMPORDPSMtoR( x86SSERegType to, u32 from )        { CMPPSMtoR( 7 ); }
void CMPORDPSRtoR( x86SSERegType to, x86SSERegType from )   { CMPPSRtoR( 7 ); }

void CVTPI2PSMtoR( x86SSERegType to, u32 from )        { SSEMtoR( 0x2a0f, 0 ); }
void CVTPI2PSRtoR( x86SSERegType to, x86SSERegType from )   { SSERtoR( 0x2a0f ); }
void CVTPS2PIMtoR( x86SSERegType to, u32 from )        { SSEMtoR( 0x2d0f, 0 ); }
void CVTPS2PIRtoR( x86SSERegType to, x86SSERegType from )   { SSERtoR( 0x2d0f ); }

void CVTDQ2PSMtoR( x86SSERegType to, u32 from )        { SSEMtoR( 0x5b0f, 0 ); }
void CVTDQ2PSRtoR( x86SSERegType to, x86SSERegType from )   { SSERtoR( 0x5b0f ); }
void CVTPS2DQMtoR( x86SSERegType to, u32 from )        { SSEMtoR66( 0x5b0f ); }
void CVTPS2DQRtoR( x86SSERegType to, x86SSERegType from )   { SSERtoR66( 0x5b0f ); }

void MAXPSMtoR( x86SSERegType to, u32 from )           { SSEMtoR( 0x5f0f, 0 ); }
void MAXPSRtoR( x86SSERegType to, x86SSERegType from )      { SSERtoR( 0x5f0f ); }

void MINPSMtoR( x86SSERegType to, u32 from )           { SSEMtoR( 0x5d0f, 0 ); }
void MINPSRtoR( x86SSERegType to, x86SSERegType from )      { SSERtoR( 0x5d0f ); }

void RSQRTPSMtoR( x86SSERegType to, u32 from )         { SSEMtoR( 0x520f, 0 ); }
void RSQRTPSRtoR( x86SSERegType to, x86SSERegType from )    { SSERtoR( 0x520f ); }

void SQRTPSMtoR( x86SSERegType to, u32 from )          { SSEMtoR( 0x510f, 0 ); }
void SQRTPSRtoR( x86SSERegType to, x86SSERegType from )     { SSERtoR( 0x510f ); }

void PMAXSWRtoR( x86SSERegType to, x86SSERegType from )		{ SSERtoR( 0xEE0F ); }
void PMINSWRtoR( x86SSERegType to, x86SSERegType from )     { SSERtoR( 0xEA0F ); }

void UNPCKLPSRtoR( x86SSERegType to, x86SSERegType from )	{ SSERtoR( 0x140F ); }

void SHUFPSRtoR( x86SSERegType to, x86SSERegType from, u8 imm8 )	{ SSERtoR( 0xC60F ); write8( imm8 ); }
void SHUFPSMtoR( x86SSERegType to, u32 from, u8 imm8 )		{ SSEMtoR( 0xC60F, 1 ); write8( imm8 ); }
